package at.jku.ssw.map.test;

import java.util.Collection;
import java.util.HashMap;
import java.util.Set;

import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import at.jku.ssw.Iterator;
import at.jku.ssw.map.Map;
import at.jku.ssw.map.Pair;

public abstract class MapTest {

	abstract Map createMap();

	public void testContentAgainstHashMap(HashMap<Object, Object> shouldBe, Map is) {
		Assert.assertEquals("Map impl not of same size", shouldBe.size(), is.size());
		/*
		 * We cannot test the order or something but we can test that, based on
		 * the iterators, there are no duplicates or missing values;
		 */

		Set<Object> shouldBeKeys = shouldBe.keySet();
		Collection<Object> shouldBeValues = shouldBe.values();

		Iterator it = is.iterator();
		while (it.hasNext()) {
			Pair p = (Pair) it.next();
			Assert.assertTrue("Key " + p.key + " was not in reference map", shouldBeKeys.contains(p.key));
			Assert.assertTrue("Value " + p.value + " was not in reference map", shouldBeValues.contains(p.value));
		}

		for (java.util.Map.Entry<Object, Object> shouldBeEntry : shouldBe.entrySet()) {
			Assert.assertTrue("Our map does not contain key " + shouldBeEntry.getKey(),
					is.containsKey(shouldBeEntry.getKey()));

			Object val = is.get(shouldBeEntry.getKey());
			Assert.assertEquals("Values do not match", shouldBeEntry.getValue(), val);
		}

	}

	@Test
	public void testSingle() {
		Object[] valKeys = { "david", new Integer(1) };
		testContentAgainstHashMap(buildReference(valKeys), buildConcrete(createMap(), valKeys));
	}

	@Rule
	public ExpectedException thrown = ExpectedException.none();

	@Test
	public void testNullKey() {
		Map map = createMap();
		thrown.expect(NullPointerException.class);
		map.put(null, new Object());
	}

	@Test
	public void testNullValue() {
		Map map = createMap();
		Object key = new Object();
		map.put(key, null);
		Assert.assertTrue(map.containsKey(key));
		Assert.assertEquals(null, map.get(key));
	}

	@Test
	public void testNotContaining() {
		Map map = createMap();
		Object key = new Object();
		Assert.assertFalse(map.containsKey(key));
		Assert.assertEquals(null, map.get(key));
	}

	@Test
	public void testMultipleNoCollision() {
		Object[] valKeys = { //
				"a", new Integer(1), //
				"b", new Integer(1), //
				"c", new Integer(1), //
				"d", new Integer(1), //
				"e", new Integer(1), //
				"f", new Integer(1), //
				"g", new Integer(1), //
				"h", new Integer(1), //
				"i", new Integer(1), //
				"j", new Integer(1), //
				"k", new Integer(1), //
				"l", new Integer(1)//
		};
		testContentAgainstHashMap(buildReference(valKeys), buildConcrete(createMap(), valKeys));
	}

	@Test
	public void testMultipleAlwaysCollide() {
		Object[] valKeys = { //
				new AlwaysCollide("a"), new Integer(1), //
				new AlwaysCollide("b"), new Integer(2), //
				new AlwaysCollide("c"), new Integer(3), //
				new AlwaysCollide("d"), new Integer(4), //
				new AlwaysCollide("e"), new Integer(5), //
				new AlwaysCollide("f"), new Integer(7), //
				new AlwaysCollide("g"), new Integer(8), //
				new AlwaysCollide("h"), new Integer(9), //
				new AlwaysCollide("i"), new Integer(10), //
				new AlwaysCollide("j"), new Integer(11), //
				new AlwaysCollide("k"), new Integer(12), //
				new AlwaysCollide("l"), new Integer(13)//
		};
		testContentAgainstHashMap(buildReference(valKeys), buildConcrete(createMap(), valKeys));
	}

	@Test
	public void testMultipleAlwaysCollideSearch() {
		Object[] valKeys = { //
				new AlwaysCollide("a"), new Integer(1), //
				new AlwaysCollide("b"), new Integer(2), //
				new AlwaysCollide("c"), new Integer(3), //
				new AlwaysCollide("d"), new Integer(4), //
				new AlwaysCollide("e"), new Integer(5), //
				new AlwaysCollide("f"), new Integer(7), //
				new AlwaysCollide("g"), new Integer(8), //
				new AlwaysCollide("h"), new Integer(9), //
				new AlwaysCollide("i"), new Integer(10), //
				new AlwaysCollide("j"), new Integer(11), //
				new AlwaysCollide("k"), new Integer(12), //
				new AlwaysCollide("l"), new Integer(13)//
		};
		HashMap<Object, Object> reference = buildReference(valKeys);
		Map concrete = buildConcrete(createMap(), valKeys);
		testContentAgainstHashMap(reference, concrete);
		// remove first element
		Object key = valKeys[0];
		Object value = valKeys[1];
		Assert.assertEquals(value, concrete.remove(key));
		reference.remove(key);
		testContentAgainstHashMap(reference, concrete);
		// search for last element
		key = valKeys[valKeys.length - 2];
		value = valKeys[valKeys.length - 1];
		Assert.assertTrue(concrete.containsKey(key));
		Assert.assertEquals(value, concrete.remove(key));
		reference.remove(key);
		testContentAgainstHashMap(reference, concrete);
	}

	@Test
	public void testMultipleAlwaysCollideMultiInsert() {
		Object[] valKeys = { //
				new AlwaysCollide("a"), new Integer(1), //
				new AlwaysCollide("b"), new Integer(2), //
				new AlwaysCollide("c"), new Integer(3), //
				new AlwaysCollide("d"), new Integer(4), //
				new AlwaysCollide("e"), new Integer(5), //
				new AlwaysCollide("f"), new Integer(7), //
				new AlwaysCollide("g"), new Integer(8), //
				new AlwaysCollide("h"), new Integer(9), //
				new AlwaysCollide("i"), new Integer(10), //
				new AlwaysCollide("j"), new Integer(11), //
				new AlwaysCollide("k"), new Integer(12), //
				new AlwaysCollide("l"), new Integer(13)//
		};
		HashMap<Object, Object> reference = buildReference(valKeys);
		Map concrete = buildConcrete(createMap(), valKeys);
		testContentAgainstHashMap(reference, concrete);
		// remove first element
		Object key = valKeys[0];
		Object value = valKeys[1];
		Assert.assertEquals(value, concrete.remove(key));
		reference.remove(key);
		testContentAgainstHashMap(reference, concrete);
		// reinsert element
		key = valKeys[valKeys.length - 2];
		value = new Integer(345);
		concrete.put(key, value);
		reference.put(key, value);
		Assert.assertTrue(concrete.containsKey(key));
		Assert.assertEquals(value, concrete.remove(key));
		reference.remove(key);
		Assert.assertFalse(concrete.containsKey(key));
		testContentAgainstHashMap(reference, concrete);
	}

	@Test
	public void testReplace() {
		Map map = createMap();
		Assert.assertEquals(0, map.size());
		Object key = "key";

		Object value0 = "value0";
		map.put(key, value0);
		Assert.assertEquals(1, map.size());
		Assert.assertTrue(map.containsKey(key));
		Assert.assertEquals(value0, map.get(key));

		Object value1 = "value1";
		map.put(key, value1);
		Assert.assertEquals(1, map.size());
		Assert.assertTrue(map.containsKey(key));
		Assert.assertEquals(value1, map.get(key));

		Object value2 = "value2";
		Assert.assertEquals(value1, map.remove(key));
		Assert.assertEquals(0, map.size());
		map.put(key, value2);
		Assert.assertEquals(1, map.size());
		Assert.assertTrue(map.containsKey(key));
		Assert.assertEquals(value2, map.get(key));
	}

	@Test
	public void testRemoveEmpty() {
		Map map = createMap();
		Assert.assertEquals(0, map.size());
		Object key = "key";
		Assert.assertFalse(map.containsKey(key));
		Object e = map.remove(key);
		Assert.assertEquals(null, e);
	}

	@Test
	public void testMultipleAlwaysCollideRemove() {
		Object[] valKeys = { //
				new AlwaysCollide("a"), new Integer(1), //
				new AlwaysCollide("b"), new Integer(2), //
				new AlwaysCollide("c"), new Integer(3), //
				new AlwaysCollide("d"), new Integer(4), //
				new AlwaysCollide("e"), new Integer(5), //
				new AlwaysCollide("f"), new Integer(7) //
		};
		HashMap<Object, Object> reference = buildReference(valKeys);
		Map concrete = buildConcrete(createMap(), valKeys);
		testContentAgainstHashMap(reference, concrete);

		Object key = valKeys[0];
		Object value = valKeys[1];

		Object e = concrete.remove(key);
		Assert.assertEquals(value, e);
		reference.remove(key);
		testContentAgainstHashMap(reference, concrete);
	}

	private static class AlwaysCollide {

		String name;

		AlwaysCollide(String name) {
			this.name = name;
		}

		@Override
		public int hashCode() {
			return 42/* the answer to everything */;
		}

		@Override
		public String toString() {
			return name;
		}
	}

	private static HashMap<Object, Object> buildReference(Object... objects) {
		HashMap<Object, Object> m = new HashMap<>();
		for (int i = 0; i < objects.length; i += 2) {
			m.put(objects[i], objects[i + 1]);
		}
		return m;
	}

	private static Map buildConcrete(Map container, Object... objects) {
		for (int i = 0; i < objects.length; i += 2) {
			container.put(objects[i], objects[i + 1]);
		}
		return container;
	}

}
